#include "scalespace.h"
#include "image_math.h"
#include "image_io.h"

#include <cmath>

using namespace stdi;
using namespace scalespace;

#define NPDE 50
#define NREMOVE 10


#define SCALESPACE_PRECISION 4

float *scalespace::gaussianKernel(float scale, int &w, int deg) 
{
  w = max(2, (int) (scale*max(1, min(10, SCALESPACE_PRECISION))));
  float *kernel = new  float[2*w+1]; 
  if (!kernel) {
    cerr << "Out of memory" << endl;
    exit(1);
  }
  float *ker = kernel + w;
  vector<float> G = dog(deg, scale);
  for(int i = -w; i <= w; i++) 
    ker[i] = polyval(G, (float)i) * std::exp(-(float)(i*i)/(2.0*scale*scale));

  return kernel;
}

// Assume that the n-order derivative of 1-D Gaussian kernel
// is $ P(x)exp(\frac{-x^2}{2*\sigma^2} $.
// This function compute the polynomial $P(x)$.
vector<float> scalespace::dog(int m, float sigma)
{
  //  const float S2(sigma*sigma);
  //  const float PI(3.1415927);
  //  float a, b;
  const double S2 = sigma*sigma;
  const double PI = 3.1415927;
  double a, b;

  int i, deg;
  vector<float> G(m+1);
  vector<float> G1(m+1); 
  for (i = 0; i <= m; i++) G[i] = 0;
  G[0] =  sqrt(1/(2*PI*S2));
  for(deg = 1; deg <= m; deg++){
    for ( i = 0; i <= m; i++){
      if ( (i+1) >= 0 && (i+1) <= m )   a = G[i+1]; 
      else a = 0;      
      if ( (i-1) >=0 && (i-1) <= m )  b = G[i-1]; 
      else b = 0;      
      G1[i] = (i+1)*a + (-1/S2) * b;      
    }
    for ( i = 0; i <= m; i++)   G[i] = G1[i];
  }
  return G;
}
 
float scalespace::polyval(const vector<float> &p, float x)
{
  float s(0), xpow(1), n(p.size());
  for (int i=0; i<n; i++, xpow *= x)  s += p[i]*xpow;
  return s;
}


// 1D smoothing
void __smooth_1d__(float *in, float *out, int n, float *ker, int w)  {
        
  int i, j, i_j;
  float s, t;

  for(i = 0; i < n; i++) {
    s = 0; t = 0;
    for(j = -w, i_j = i+w; j <= w; j++, i_j--)
      if (i_j >= 0 && i_j < n) 
	s  += in[i_j] * ker[j];
      else t += ker[j];
    out[i] = s/(1-t);
  }

}



FloatImage scalespace::smooth(const Image &img, float scale, const float ratio){
    
  const int nx = img.nx();
  const int ny = img.ny();
  const int nz = img.nz();
  const int npix = img.npix();
  
  //  FloatImage I(nx, ny, nz);
  //  for(int i=0; i<npix; i++) I.set(i, img.pixel(i));
  FloatImage I = toFloatImage(img);
  
  int x, y, z;
  float *in, *out; 
  int w;
  float * kernel = gaussianKernel(scale, w, 0);
  float * ker = kernel + w;

  int zw;
  float zscale = scale/ratio;
  float * zkernel = gaussianKernel(zscale, zw, 0);
  float * zker = zkernel + zw;
  
  // x direction
  if (nx>1) {
    //cout << "x...";
    in = new  float[nx]; 
    out = new  float[nx];
    if (!in || !out) {
      cerr << "Out of memory" << endl;
      exit(1);
    }

    for(z=0; z<nz; z++)
      for(y=0; y<ny; y++) {
	I.lineX(y, z, in);
	__smooth_1d__(in, out, nx, ker, w);
	I.setLineX(y, z, out);
      }
    delete []in; delete []out;
  }

  // y direction
  if (ny>1) {
    //cout << "y...";
    in = new  float[ny]; 
    out = new  float[ny];
    if (!in || !out) {
      cerr << "Out of memory" << endl;
      exit(1);
    }
    for(z=0; z<nz; z++)
      for(x=0; x<nx; x++)  {
	I.lineY(x, z, in);
	__smooth_1d__(in, out, ny, ker, w);
	I.setLineY(x, z, out);
      }
    delete []in; delete []out;
  }

  // z direction
  if (nz>1) {
    //cout << "z...";   
    in = new  float[nz]; 
    out = new  float[nz];
    if (!in || !out) {
      cerr << "Out of memory" << endl;
      exit(1);
    }

    for(y=0; y<ny; y++)
      for(x=0; x<nx; x++) {
	I.lineZ(x, y, in);
	__smooth_1d__(in, out, nz, zker, zw);
	I.setLineZ(x, y, out);
      }
    delete []in; delete []out;
  }
 
  delete []kernel;
  delete []zkernel; 
  //  cout << "done" << endl;  
  return I;
}


FloatImage scalespace::smooth(const Image &img, float scale_x, float scale_y, float scale_z){
    
  const int nx = img.nx();
  const int ny = img.ny();
  const int nz = img.nz();
  const int npix = img.npix();
  
  //  FloatImage I(nx, ny, nz);
  //  for(int i=0; i<npix; i++) I.set(i, img.pixel(i));
  FloatImage I = toFloatImage(img);
  
  int x, y, z;
  float *in, *out; 
  int xw;
  float * kernel_x = gaussianKernel(scale_x, xw, 0);
  float * ker_x = kernel_x + xw;

  int yw;
  float * kernel_y = gaussianKernel(scale_y, yw, 0);
  float * ker_y = kernel_y + yw;

  int zw;
  float * kernel_z = gaussianKernel(scale_z, zw, 0);
  float * ker_z = kernel_z + zw;
  
  // x direction
  if (nx>1) {
    //cout << "x...";
    in = new  float[nx]; 
    out = new  float[nx];
    if (!in || !out) {
      cerr << "Out of memory" << endl;
      exit(1);
    }

    for(z=0; z<nz; z++)
      for(y=0; y<ny; y++) {
	I.lineX(y, z, in);
	__smooth_1d__(in, out, nx, ker_x, xw);
	I.setLineX(y, z, out);
      }
    delete []in; delete []out;
  }

  // y direction
  if (ny>1) {
    //cout << "y...";
    in = new  float[ny]; 
    out = new  float[ny];
    if (!in || !out) {
      cerr << "Out of memory" << endl;
      exit(1);
    }
    for(z=0; z<nz; z++)
      for(x=0; x<nx; x++)  {
	I.lineY(x, z, in);
	__smooth_1d__(in, out, ny, ker_y, yw);
	I.setLineY(x, z, out);
      }
    delete []in; delete []out;
  }

  // z direction
  if (nz>1) {
    //cout << "z...";   
    in = new  float[nz]; 
    out = new  float[nz];
    if (!in || !out) {
      cerr << "Out of memory" << endl;
      exit(1);
    }

    for(y=0; y<ny; y++)
      for(x=0; x<nx; x++) {
	I.lineZ(x, y, in);
	__smooth_1d__(in, out, nz, ker_z, zw);
	I.setLineZ(x, y, out);
      }
    delete []in; delete []out;
  }
 
  delete []kernel_x;
  delete []kernel_y;
  delete []kernel_z;
  
  //  cout << "done" << endl;  
  return I;
}
  

ByteImage __features_3d__(const Image &img, float scale, FeatureType ft, const float ratio)   {
  
  if (img.isNull()) return ByteImage();
  FloatImage pimg = scalespace::smooth(img, scale, ratio);
  if (pimg.isNull()) return ByteImage();
  
  const int nx = pimg.nx();
  const int ny = pimg.ny();
  const int nz = pimg.nz();       
  const int nx_1 = pimg.nx()-1;
  const int ny_1 = pimg.ny()-1;
  const int nz_1 = pimg.nz()-1;   
  
  int z, x, y;  
  float Ix, Iy, Iz, Ixx, Iyy, Izz, Ixy, Ixz, Iyz, det, t, le;
  bool ef2;
  ByteImage feat(nx, ny, nz);
  feat.reset(byte(Background));
  
  
  switch (ft) {
  case BrightZC:
    for(z = 1; z < nz_1; z++) 
      for(y = 1; y < ny_1; y++)
	for(x = 1; x < nx_1; x++) {      	  
	  t = 2.0 * pimg.at(x,y,z);
	  Ixx =  pimg.at(x+1, y, z) + pimg.at(x-1, y, z) - t; 
	  Iyy =  pimg.at(x, y+1, z) + pimg.at(x, y-1, z) - t; 
	  Izz =  pimg.at(x, y, z+1) + pimg.at(x, y, z-1) - t; 
	  if (Ixx+Iyy+Izz<0) feat.set(x,y,z, byte(Foreground));	    
	}
    break;
  case DarkZC:
    for(z = 1; z < nz_1; z++) 
      for(y = 1; y < ny_1; y++)
	for(x = 1; x < nx_1; x++) {      	  
	  t = 2.0 * pimg.at(x,y,z);
	  Ixx =  pimg.at(x+1, y, z) + pimg.at(x-1, y, z) - t; 
	  Iyy =  pimg.at(x, y+1, z) + pimg.at(x, y-1, z) - t; 
	  Izz =  pimg.at(x, y, z+1) + pimg.at(x, y, z-1) - t; 
	  if (Ixx+Iyy+Izz>0) feat.set(x,y,z, byte(Foreground));	    
	}
    break;
  case BrightLE:
    for(z = 1; z < nz_1; z++) 
      for(y = 1; y < ny_1; y++)
	for(x = 1; x < nx_1; x++) {      	  
	  t = 2.0 * pimg.at(x,y,z);
	  Ixx =  pimg.at(x+1, y, z) + pimg.at(x-1, y, z) - t; 
	  Iyy =  pimg.at(x, y+1, z) + pimg.at(x, y-1, z) - t; 
	  Izz =  pimg.at(x, y, z+1) + pimg.at(x, y, z-1) - t; 
	  Ixy = 0.25 * (pimg.at(x+1, y+1, z) + pimg.at(x-1, y-1, z) -
			pimg.at(x+1, y-1, z) - pimg.at(x-1, y+1, z));
	  Ixz = 0.25 * (pimg.at(x+1, y, z+1) + pimg.at(x-1, y, z-1) -
			pimg.at(x-1, y, z+1) - pimg.at(x+1, y, z-1));
	  Iyz = 0.25 * (pimg.at(x, y+1, z+1) + pimg.at(x, y-1, z-1) -
			pimg.at(x, y-1, z+1) - pimg.at(x, y+1, z-1));
	  
	  Ix =  0.5 *( pimg.at(x+1, y, z) - pimg.at(x-1, y, z) ); 
	  Iy =  0.5 *(pimg.at(x, y+1, z) - pimg.at(x, y-1, z) ); 
	  Iz =  0.5 *(pimg.at(x, y, z+1) - pimg.at(x, y, z-1) ); 
	  
	  le = Ix*Ix*Ixx + Iy*Iy*Iyy + Iz*Iz*Izz + 2 * (Ix*Iy*Ixy + Ix*Iz*Ixz + Iy*Iz*Iyz );
	  if (le < 0) feat.set(x, y, z, byte(Foreground));	    
	}
    break;
  case DarkLE:
    for(z = 1; z < nz_1; z++) 
      for(y = 1; y < ny_1; y++)
	for(x = 1; x < nx_1; x++) {      	  
	  t = 2.0 * pimg.at(x,y,z);
	  Ixx =  pimg.at(x+1, y, z) + pimg.at(x-1, y, z) - t; 
	  Iyy =  pimg.at(x, y+1, z) + pimg.at(x, y-1, z) - t; 
	  Izz =  pimg.at(x, y, z+1) + pimg.at(x, y, z-1) - t; 
	  Ixy = 0.25 * (pimg.at(x+1, y+1, z) + pimg.at(x-1, y-1, z) -
			pimg.at(x+1, y-1, z) - pimg.at(x-1, y+1, z));
	  Ixz = 0.25 * (pimg.at(x+1, y, z+1) + pimg.at(x-1, y, z-1) -
			pimg.at(x-1, y, z+1) - pimg.at(x+1, y, z-1));
	  Iyz = 0.25 * (pimg.at(x, y+1, z+1) + pimg.at(x, y-1, z-1) -
			pimg.at(x, y-1, z+1) - pimg.at(x, y+1, z-1));
	  
	  Ix =  0.5 *( pimg.at(x+1, y, z) - pimg.at(x-1, y, z) ); 
	  Iy =  0.5 *(pimg.at(x, y+1, z) - pimg.at(x, y-1, z) ); 
	  Iz =  0.5 *(pimg.at(x, y, z+1) - pimg.at(x, y, z-1) ); 
	  
	  le = Ix*Ix*Ixx + Iy*Iy*Iyy + Iz*Iz*Izz + 2 * (Ix*Iy*Ixy + Ix*Iz*Ixz + Iy*Iz*Iyz );
	  if (le > 0) feat.set(x, y, z, byte(Foreground));	    
	}
    break;      
    
    
  case BrightEF:
    for(z = 1; z < nz_1; z++) 
      for(y = 1; y < ny_1; y++)
	for(x = 1; x < nx_1; x++) {      	  
	  t = 2.0 * pimg.at(x,y,z);
	  Ixx =  pimg.at(x+1, y, z) + pimg.at(x-1, y, z) - t; 
	  Iyy =  pimg.at(x, y+1, z) + pimg.at(x, y-1, z) - t; 
	  Izz =  pimg.at(x, y, z+1) + pimg.at(x, y, z-1) - t; 
	  Ixy = 0.25 * (pimg.at(x+1, y+1, z) + pimg.at(x-1, y-1, z) -
			pimg.at(x+1, y-1, z) - pimg.at(x-1, y+1, z));
	  Ixz = 0.25 * (pimg.at(x+1, y, z+1) + pimg.at(x-1, y, z-1) -
			pimg.at(x-1, y, z+1) - pimg.at(x+1, y, z-1));
	  Iyz = 0.25 * (pimg.at(x, y+1, z+1) + pimg.at(x, y-1, z-1) -
			pimg.at(x, y-1, z+1) - pimg.at(x, y+1, z-1));

	  det = Ixx*(Iyy*Izz-Iyz*Iyz) - Ixy*(Ixy*Izz-Ixz*Iyz) + Ixz*(Ixy*Iyz-Iyy*Ixz);
	  ef2 = Ixx*Iyy>Ixy*Ixy  && Ixx*Izz>Ixz*Ixz && Iyy*Izz>Iyz*Iyz;		 
	  if ( ef2 && Ixx < 0  && det < 0 )  feat.set(x, y, z, byte(Foreground));	    
	}
    break;
    
    
  case DarkEF:
    for(z = 1; z < nz_1; z++) 
      for(y = 1; y < ny_1; y++)
	for(x = 1; x < nx_1; x++) {      	  
	  t = 2.0 * pimg.at(x,y,z);
	  Ixx =  pimg.at(x+1, y, z) + pimg.at(x-1, y, z) - t; 
	  Iyy =  pimg.at(x, y+1, z) + pimg.at(x, y-1, z) - t; 
	  Izz =  pimg.at(x, y, z+1) + pimg.at(x, y, z-1) - t; 
	  Ixy = 0.25 * (pimg.at(x+1, y+1, z) + pimg.at(x-1, y-1, z) -
			pimg.at(x+1, y-1, z) - pimg.at(x-1, y+1, z));
	  Ixz = 0.25 * (pimg.at(x+1, y, z+1) + pimg.at(x-1, y, z-1) -
			pimg.at(x-1, y, z+1) - pimg.at(x+1, y, z-1));
	  Iyz = 0.25 * (pimg.at(x, y+1, z+1) + pimg.at(x, y-1, z-1) -
			pimg.at(x, y-1, z+1) - pimg.at(x, y+1, z-1));
	  
	  det = Ixx*(Iyy*Izz-Iyz*Iyz) - Ixy*(Ixy*Izz-Ixz*Iyz) + Ixz*(Ixy*Iyz-Iyy*Ixz);
	  ef2 = Ixx*Iyy>Ixy*Ixy  && Ixx*Izz>Ixz*Ixz && Iyy*Izz>Iyz*Iyz;		 
	  if ( ef2 && Ixx > 0  && det > 0)  feat.set(x, y, z, byte(Foreground));	    
	}
    break;
    
  default: ; // Nothing to do
    
  }
  
  return feat;
}


#if 0

ByteImage __features_2d__(const Image &img, float scale, FeatureType ft)   {

  if (img.isNull()) return ByteImage();

  if (ft==BrightZC || ft==DarkZC) {
    FloatImage Ixx = dXX(img, scale);
    FloatImage Iyy = dYY(img, scale);
    ByteImage e(img.nx(), img.ny());
    if (ft==BrightZC)
      for(int i=0; i<e.npix(); i++)
	e(i) = Ixx(i)+Iyy(i) < 0? Foreground:Background;
    else 
      for(int i=0; i<e.npix(); i++)
	e(i) = Ixx(i)+Iyy(i) > 0? Foreground:Background;

    return e;
  }


  FloatImage pimg = scalespace::smooth(img, scale);
  const int nx = pimg.nx();
  const int ny = pimg.ny();
  const int nx_1 = pimg.nx()-1;
  const int ny_1 = pimg.ny()-1;
  
  int x, y;  
  float Ix, Iy, Ixx, Iyy, Ixy, t, le, ef2;
  ByteImage feat(nx, ny);
  if (Background != 0)
    feat.reset(byte(Background));

  switch (ft) {
  case BrightZC:
    for(y = 1; y < ny_1; y++)
      for(x = 1; x < nx_1; x++) {      	  
	t = 2.0 * pimg.at(x,y);
	Ixx =  pimg.at(x+1, y) + pimg.at(x-1, y) - t; 
	Iyy =  pimg.at(x, y+1) + pimg.at(x, y-1) - t; 
	if (Ixx+Iyy<0) feat.set(x, y, byte(Foreground));	    
      }
     return feat;
  case DarkZC:
    for(y = 1; y < ny_1; y++)
      for(x = 1; x < nx_1; x++) {      	  
	t = 2.0 * pimg.at(x, y);
	Ixx =  pimg.at(x+1, y) + pimg.at(x-1, y) - t; 
	Iyy =  pimg.at(x, y+1) + pimg.at(x, y-1) - t; 
	if (Ixx+Iyy>0) feat.set(x, y, byte(Foreground));	    
      }
     return feat;
  case BrightLE:
    for(y = 1; y < ny_1; y++)
      for(x = 1; x < nx_1; x++) {      	  
	t = 2.0 * pimg.at(x, y);
	Ixx =  pimg.at(x+1, y) + pimg.at(x-1, y) - t; 
	Iyy =  pimg.at(x, y+1) + pimg.at(x, y-1) - t; 
	Ixy = 0.25 * (pimg.at(x+1, y+1) + pimg.at(x-1, y-1) -
		      pimg.at(x+1, y-1) - pimg.at(x-1, y+1));
	
	Ix =  0.5 *( pimg.at(x+1, y) - pimg.at(x-1, y) ); 
	Iy =  0.5 *(pimg.at(x, y+1) - pimg.at(x, y-1) ); 
	
	if (Ix*Ix*Ixx+Iy*Iy*Iyy+2*Ix*Iy*Ixy < 0) 
	  feat.set(x, y, byte(Foreground));	    
      }
     return feat;
  case DarkLE:
    for(y = 1; y < ny_1; y++)
      for(x = 1; x < nx_1; x++) {      	  
	t = 2.0 * pimg.at(x,y);
	Ixx =  pimg.at(x+1, y) + pimg.at(x-1, y) - t; 
	Iyy =  pimg.at(x, y+1) + pimg.at(x, y-1) - t; 
	Ixy = 0.25 * (pimg.at(x+1, y+1) + pimg.at(x-1, y-1) -
		      pimg.at(x+1, y-1) - pimg.at(x-1, y+1));
	
	Ix =  0.5 *( pimg.at(x+1, y) - pimg.at(x-1, y) ); 
	Iy =  0.5 *(pimg.at(x, y+1) - pimg.at(x, y-1) ); 	  
	if (Ix*Ix*Ixx+Iy*Iy*Iyy+2*Ix*Iy*Ixy > 0) 
	  feat.set(x, y, byte(Foreground));	    
      }
     return feat;    
    
  case BrightEF:
    for(y = 1; y < ny_1; y++)
      for(x = 1; x < nx_1; x++) {      	  
	t = 2.0 * pimg.at(x,y);
	Ixx =  pimg.at(x+1, y) + pimg.at(x-1, y) - t; 
	Iyy =  pimg.at(x, y+1) + pimg.at(x, y-1) - t; 
	Ixy = 0.25 * (pimg.at(x+1, y+1) + pimg.at(x-1, y-1) -
		      pimg.at(x+1, y-1) - pimg.at(x-1, y+1));
	
	if ( Ixx*Iyy>Ixy*Ixy && Ixx < 0)  feat.set(x, y, byte(Foreground));	    
      }

    return feat;
  case DarkEF:
    
    for(y = 1; y < ny_1; y++)
      for(x = 1; x < nx_1; x++) {      	  
	t = 2.0 * pimg.at(x,y);
	Ixx =  pimg.at(x+1, y) + pimg.at(x-1, y) - t; 
	Iyy =  pimg.at(x, y+1) + pimg.at(x, y-1) - t; 
	Ixy = 0.25 * (pimg.at(x+1, y+1) + pimg.at(x-1, y-1) -
		      pimg.at(x+1, y-1) - pimg.at(x-1, y+1));
	
	if ( Ixx*Iyy>Ixy*Ixy && Ixx > 0)  
	  feat.set(x, y, byte(Foreground));	    
      }
    return feat;
    
  default: ; // Nothing to do
    
  }

  return feat;
}

#endif



ByteImage __features_2d__(const Image &img, float scale, FeatureType ft)   
{

  if (img.isNull()) return ByteImage();

  if (ft==BrightZC || ft==DarkZC) {
    FloatImage Ixx = dXX(img, scale);
    FloatImage Iyy = dYY(img, scale);
    ByteImage e(img.nx(), img.ny());
    if (ft==BrightZC)
      for(int i=0; i<e.npix(); i++)
	e(i) = Ixx(i)+Iyy(i) < 0? Foreground:Background;
    else 
      for(int i=0; i<e.npix(); i++)
	e(i) = Ixx(i)+Iyy(i) > 0? Foreground:Background;

    return e;
  }

  FloatImage I = scalespace::smooth(img, scale);
  const int nx = I.nx();
  const int ny = I.ny();
  const int nx_1 = I.nx()-1;
  const int ny_1 = I.ny()-1;
  const int nx1 = I.nx()+1;
  
  int x, y, k;  
  float Ix, Iy, Ixx, Iyy, Ixy, t;
  ByteImage feat(nx, ny);
  
  float *p = I.ptr();
  float *q;

  switch (ft) {
  case BrightZC:
    for(y = 1; y < ny_1; y++)
      for(x = 1; x < nx_1; x++) {
	k = x+y*nx; q = p+k; 
	t = 2.0 * *q;
	Ixx =  q[1] + q[-1] - t; 
	Iyy =  q[nx] + q[-nx] - t; 
	if (Ixx+Iyy<0) 
	  feat(k) = Foreground;	    
      }
    break;
  case DarkZC:
    for(y = 1; y < ny_1; y++)
      for(x = 1; x < nx_1; x++) { 
	k = x+y*nx;
	q = p+k;
     	  
	t = 2.0 * *q;
	Ixx =  q[1] + q[-1] - t; 
	Iyy =  q[nx] + q[-nx] - t; 
	if (Ixx+Iyy>0) 
	  feat(k) = Foreground;	    
      }
     break;
  case BrightLE:
    for(y = 1; y < ny_1; y++)
      for(x = 1; x < nx_1; x++) {      	  
	k = x+y*nx; 
	q = p+k; 
	t = 2.0 * *q;
	Ixx =  q[1] + q[-1] - t; 
	Iyy =  q[nx] + q[-nx] - t; 
	Ixy = 0.25 * (q[nx1] + q[-nx1] - q[-nx_1] - q[nx_1]);
	
	Ix =  0.5 *( q[1] - q[-1] ); 
	Iy =  0.5 *(q[nx] - q[-nx] ); 
	
	if (Ix*Ix*Ixx+Iy*Iy*Iyy+2*Ix*Iy*Ixy < 0) 
	  feat(k) = Foreground;	    
      }
     break;
  case DarkLE:
    for(y = 1; y < ny_1; y++)
      for(x = 1; x < nx_1; x++) {      	  
	k = x+y*nx; 
	q = p+k; 
	t = 2.0 * *q;
	Ixx =  q[1] + q[-1] - t; 
	Iyy =  q[nx] + q[-nx] - t; 
	Ixy = 0.25 * (q[nx1] + q[-nx1] - q[-nx_1] - q[nx_1]);
	
	Ix =  0.5 *( q[1] - q[-1] ); 
	Iy =  0.5 *(q[nx] - q[-nx] ); 	  
	if (Ix*Ix*Ixx+Iy*Iy*Iyy+2*Ix*Iy*Ixy > 0) 
	  feat(k) = Foreground;  
      }
     break;    
    
  case BrightEF:
    for(y = 1; y < ny_1; y++)
      for(x = 1; x < nx_1; x++) {      	  

	k = x+y*nx; q = p+k; 
	t = 2.0 * *q;
	Ixx =  q[1] + q[-1] - t; 
	Iyy =  q[nx] + q[-nx] - t; 
	Ixy = 0.25 * (q[nx1] + q[-nx1] - q[-nx_1] - q[nx_1]);	
	if ( Ixx*Iyy>Ixy*Ixy && Ixx < 0)  
	  feat(k) = Foreground;	    
      }
    
    break;
  case DarkEF:
    
    for(y = 1; y < ny_1; y++)
      for(x = 1; x < nx_1; x++) {      	  
	
	k = x+y*nx; 
	q = p+k; 
	
	t = 2.0 * *q;
	Ixx =  q[1] + q[-1] - t; 
	Iyy =  q[nx] + q[-nx] - t; 
	Ixy = 0.25 * (q[nx1] + q[-nx1] - q[-nx_1] - q[nx_1]);
	
	if ( Ixx*Iyy>Ixy*Ixy && Ixx > 0)  
	  feat(k) = Foreground;	    
      }
    break;
    
  default: ; // TODO    
  }

  return feat;
}



ByteImage scalespace::features(const Image &img, float scale,  FeatureType ft, const float ratio)
{
  if (img.isNull()) return ByteImage();
  if (img.nz()>1) return __features_3d__(img, scale, ft, ratio);
  return __features_2d__(img, scale, ft);  
}


void __gaussian_1d__(float *in, float *out, int n, float *ker, int w)  {
        
  int i, j, i_j;
  float s;  

  for(i = 0; i < n; i++) {
    s = 0; 
    for(j = -w, i_j = i+w; j <= w; j++, i_j--)
      if (i_j < 0)
	s  += in[0] * ker[j];
      else if (i_j>=n)
	s  += in[n-1] * ker[j];
      else
	s  += in[i_j] * ker[j];    
    out[i] = s;
  }

}


FloatImage scalespace::gaussian(const Image &img, float scale,
				int order_x, int order_y, int order_z)
{    
  const int nx = img.nx();
  const int ny = img.ny();
  const int nz = img.nz();
  const int npix = img.npix();
  
  //FloatImage I(nx, ny, nz);
  //for(int i=0; i<npix; i++) I.set(i, img.pixel(i));
  FloatImage I = toFloatImage(img);

  int x, y, z, w;
  float *in, *out, *kernel, *ker; 

  // x direction
  if (nx>1) {
    // cout << "x...";
    in = new  float[nx]; 
    out = new  float[nx];
    if (!in || !out) {
      cerr << "Out of memory" << endl;
      exit(1);
    }

    kernel = gaussianKernel(scale, w, order_x);
    ker = kernel + w;
    for(z=0; z<nz; z++)
      for(y=0; y<ny; y++) {
	I.lineX(y, z, in);
	__gaussian_1d__(in, out, nx, ker, w);
	I.setLineX(y, z, out);
      }
    delete []in; delete []out; delete []kernel;    
  }

  // y direction
  if (ny>1 ) {
    //cout << "y...";
    in = new  float[ny]; 
    out = new  float[ny];
    if (!in || !out) {
      cerr << "Out of memory" << endl;
      exit(1);
    }

    kernel = gaussianKernel(scale, w, order_y);
    ker = kernel + w;
    for(z=0; z<nz; z++)
      for(x=0; x<nx; x++)  {
	I.lineY(x, z, in);
	__gaussian_1d__(in, out, ny, ker, w);
	I.setLineY(x, z, out);
      }
    delete []in; delete []out; delete []kernel;
  }

  // z direction
  if (nz>1) {
    //cout << "z...";   
    in = new  float[nz]; 
    out = new  float[nz];
    if (!in || !out) {
      cerr << "Out of memory" << endl;
      exit(1);
    }

    kernel = gaussianKernel(scale, w, order_z);
    ker = kernel + w;
    for(y=0; y<ny; y++)
      for(x=0; x<nx; x++) {
	I.lineZ(x, y, in);
	__gaussian_1d__(in, out, nz, ker, w);
	I.setLineZ(x, y, out);
      }
    delete []in; delete []out; delete []kernel;
  }

  //  cout << "done" << endl;  
  return I;
}

FloatImage scalespace::gaussian(const Image &img, float scale_x, float scale_y, float scale_z,
				int order_x, int order_y, int order_z)
{    
  const int nx = img.nx();
  const int ny = img.ny();
  const int nz = img.nz();
  const int npix = img.npix();
  
  //FloatImage I(nx, ny, nz);
  //for(int i=0; i<npix; i++) I.set(i, img.pixel(i));
  FloatImage I = toFloatImage(img);

  int x, y, z, w;
  float *in, *out, *kernel, *ker; 

  // x direction
  if (nx>1) {
    // cout << "x...";
    in = new  float[nx]; 
    out = new  float[nx];
    if (!in || !out) {
      cerr << "Out of memory" << endl;
      exit(1);
    }

    kernel = gaussianKernel(scale_x, w, order_x);
    ker = kernel + w;
    for(z=0; z<nz; z++)
      for(y=0; y<ny; y++) {
	I.lineX(y, z, in);
	__gaussian_1d__(in, out, nx, ker, w);
	I.setLineX(y, z, out);
      }
    delete []in; delete []out; delete []kernel;    
  }

  // y direction
  if (ny>1 ) {
    //cout << "y...";
    in = new  float[ny]; 
    out = new  float[ny];
    if (!in || !out) {
      cerr << "Out of memory" << endl;
      exit(1);
    }

    kernel = gaussianKernel(scale_y, w, order_y);
    ker = kernel + w;
    for(z=0; z<nz; z++)
      for(x=0; x<nx; x++)  {
	I.lineY(x, z, in);
	__gaussian_1d__(in, out, ny, ker, w);
	I.setLineY(x, z, out);
      }
    delete []in; delete []out; delete []kernel;
  }

  // z direction
  if (nz>1) {
    //cout << "z...";   
    in = new  float[nz]; 
    out = new  float[nz];
    if (!in || !out) {
      cerr << "Out of memory" << endl;
      exit(1);
    }

    kernel = gaussianKernel(scale_z, w, order_z);
    ker = kernel + w;
    for(y=0; y<ny; y++)
      for(x=0; x<nx; x++) {
	I.lineZ(x, y, in);
	__gaussian_1d__(in, out, nz, ker, w);
	I.setLineZ(x, y, out);
      }
    delete []in; delete []out; delete []kernel;
  }

  //  cout << "done" << endl;  
  return I;
}

// Partial derivatives
FloatImage scalespace::dX(const Image &img, float scale)
{ return gaussian(img, scale, 1, 0, 0); }

FloatImage scalespace::dY(const Image &img, float scale)
{ return gaussian(img, scale, 0, 1, 0); }

FloatImage scalespace::dZ(const Image &img, float scale)
{ return gaussian(img, scale, 0, 0, 1); }

FloatImage scalespace::dXX(const Image &img, float scale)
{ return gaussian(img, scale, 2, 0, 0); }

FloatImage scalespace::dXXX(const Image &img, float scale)
{ return gaussian(img, scale, 3, 0, 0); }

FloatImage scalespace::dYY(const Image &img, float scale)
{ return gaussian(img, scale, 0, 2, 0); }

FloatImage scalespace::dYYY(const Image &img, float scale)
{ return gaussian(img, scale, 0, 3, 0); }

FloatImage scalespace::dZZ(const Image &img, float scale)
{ return gaussian(img, scale, 0, 0, 2); }

FloatImage scalespace::dZZZ(const Image &img, float scale)
{ return gaussian(img, scale, 0, 0, 3); }

FloatImage scalespace::dXY(const Image &img, float scale)
{ return gaussian(img, scale, 1, 1, 0); }

FloatImage scalespace::dXY(const Image &img, float scale_x, float scale_y)
{ return gaussian(img, scale_x,scale_y, 1, 1, 1, 0); }

FloatImage scalespace::dXZ(const Image &img, float scale)
{ return gaussian(img, scale, 1, 0, 1); }

FloatImage scalespace::dXZ(const Image &img, float scale_x, float scale_z)
{ return gaussian(img, scale_x, 1,scale_z,  1, 0, 1); }

FloatImage scalespace::dYZ(const Image &img, float scale)
{ return gaussian(img, scale, 0, 1, 1); }

FloatImage scalespace::dYZ(const Image &img, float scale_y, float scale_z)
{ return gaussian(img, 1, scale_y, scale_z, 0, 1, 1); }

FloatImage scalespace::dXXY( const Image &img, float scale_x, float scale_y)
{ return gaussian(img, scale_x, scale_y, 1, 2, 1, 0); }

FloatImage scalespace::dXYY( const Image &img, float scale_x, float scale_y)
{ return gaussian(img, scale_x, scale_y, 1, 1, 2, 0); }

FloatImage scalespace::dXYZ( const Image &img, float scale_x, float scale_y, float scale_z)
{ return gaussian(img, scale_x, scale_y, scale_z, 1, 1, 1); }

FloatImage scalespace::dXXZ( const Image &img, float scale_x, float scale_z)
{ return gaussian(img, scale_x, 1, scale_z, 2, 0, 1); }

FloatImage scalespace::dYYZ( const Image &img, float scale_y, float scale_z)
{ return gaussian(img, 1, scale_y, scale_z, 0, 2, 1); }

FloatImage scalespace::dXZZ( const Image &img, float scale_x, float scale_z)
{ return gaussian(img, scale_x, 1, scale_z, 1, 0, 2); }

FloatImage scalespace::dYZZ( const Image &img, float scale_y, float scale_z)
{ return gaussian(img, 1, scale_y, scale_z, 0, 1, 2); }

FloatImage scalespace::gradient(const Image &img, float scale, const float ratio)
{
  FloatImage imgx = scalespace::dX(img, scale); 
  FloatImage imgy = scalespace::dY(img, scale); 
  const int npix = img.npix();
  if (img.nz()==1) {
    for(int i=0; i<npix; i++)
      imgx(i) = sqrt(imgx(i)*imgx(i) + imgy(i)*imgy(i) ); 
  } else {
    float zscale;
    if (ratio) zscale = scale/ratio;
    else zscale = scale;
    FloatImage imgz = scalespace::dZ(img, zscale); 
    for(int i=0; i<npix; i++)
      imgx(i) = sqrt(imgx(i)*imgx(i) + imgy(i)*imgy(i)+imgz(i)*imgz(i) );  
  }

  return imgx;
}


//Modified on 5/13/05 for 3D
FloatImage scalespace::laplacian(const Image &img, float scale)
{
  FloatImage imgxx = scalespace::dXX(img, scale); 
  FloatImage imgyy = scalespace::dYY(img, scale); 
  const int npix = img.npix();
  if (img.nz()==1) {
  for(int i=0; i<npix; i++)
    imgxx.set(i, imgxx.at(i) + imgyy.at(i) );
  }
  else {
    FloatImage imgzz = scalespace::dZZ(img, scale);
    for(int i=0; i<npix; i++)
      imgxx.set(i, imgxx.at(i) + imgyy.at(i) + imgzz.at(i));
  }
  return imgxx;  
}



void scalespace::laplaceEquation(FloatImage &I, const ByteImage &E)
{
  const int nx = I.nx();
  const int ny = I.ny();
  const int nx_1 = nx-1;  
  const int ny_1 = ny-1;
  const int npix = nx*ny;

  float *p = I.ptr();
  byte *q = E.ptr();
  float *r = new  float[npix];
  if (!r) {
    cerr << "Out of memory" << endl;
    exit(1);
  }

  if (!r) return;
  int  i, j, k, l, loop;
  float *s; 
  
  for(loop=0; loop<NPDE; loop++) {
    
    memcpy(r, p, npix*sizeof(float));
    for(i=1, l=nx; i<ny_1; i++, l+=nx)
      for(j=1, k=l+1; j<nx_1; j++, k++)
	if(q[k]==Foreground) {
	  s = r+k;
	  p[k] = 0.25*(*(s+1)+*(s-1)+ *(s+nx)+ *(s-nx));
	}	  
  }  
  delete []r;
}


FloatImage scalespace::harmonicCut(const Image &I0, 
				   float scale,
				   bool remove_bright_regions, 
				   int niter)
{
  const int nx = I0.nx();
  const int ny = I0.ny();
  const int npix = nx*ny;
  FloatImage I1(nx, ny);
  for(int i=0; i<npix; i++)
    I1(i) =  I0.pixel(i);

  ByteImage I2;
  
  for(int loop = 0; loop < niter; loop++) {
    I2 = scalespace::features(I1, scale,
			      remove_bright_regions ? BrightEF : DarkEF);
    laplaceEquation(I1, I2);
  }

  return I1;
}

ByteImage scalespace::hcut(const Image &I0, float scale, float zc_scale) 
{
  const int nx = I0.nx();
  const int ny = I0.ny();
  const int npix = nx*ny;
  FloatImage I1(nx, ny);
  for(int i=0; i<npix; i++) I1.set(i, I0.pixel(i));
  
  ByteImage I2;
  for(int loop=0; loop<NREMOVE; loop++) {
    I2 = scalespace::features(I1, scale, BrightEF);
    laplaceEquation(I1, I2);
  }
  
  return scalespace::features(I1, zc_scale, BrightZC);
}


ShortImage scalespace::angle(const Image &Ix, const Image &Iy)
{
  if (!Ix.equalSize(Iy)) return ShortImage();
  ShortImage a(Ix.nx(), Ix.ny(), Ix.nz());
  const int npix = Ix.npix();
  for(int i=0; i<npix; i++)
    a.set(i, (short) image_math::angle(Ix.pixel(i), Iy.pixel(i)));
  return a;
}

FloatImage scalespace::angleInRadian(const Image &Ix, const Image &Iy)
{
  if (!Ix.equalSize(Iy)) return FloatImage();
  FloatImage a(Ix.nx(), Ix.ny(), Ix.nz());
  const int npix = Ix.npix();
  for(int i=0; i<npix; i++)
    a(i) = atan2(Ix.pixel(i), Iy.pixel(i));
  return a;
}


ShortImage scalespace::gradient(const Image &Ix, const Image &Iy)
{
  if (!Ix.equalSize(Iy)) return ShortImage();
  ShortImage a(Ix.nx(), Ix.ny(), Ix.nz());
  const int npix = Ix.npix();
  for(int i=0; i<npix; i++)
    a.set(i, (short) image_math::norm(Ix.pixel(i), Iy.pixel(i)));
  return a;
}

ShortImage scalespace::gradient(const Image &Ix, const Image &Iy, const Image& Iz)
{
  if (!Ix.equalSize(Iy) || !Ix.equalSize(Iz)) return ShortImage();
  ShortImage a(Ix.nx(), Ix.ny(), Ix.nz());
  const int npix = Ix.npix();
  for(int i=0; i<npix; i++)
      a.set(i,(short) sqrt(Ix.pixel(i)*Ix.pixel(i)+Iy.pixel(i)*Iy.pixel(i)+Iz.pixel(i)*Iz.pixel(i)));
  //a.set(i, (short) image_math::norm(Ix.pixel(i), Iy.pixel(i)));
  return a;
}

ByteImage scalespace::direction(const Image &Ix, const Image &Iy)
{
  if (!Ix.equalSize(Iy)) return ByteImage();
  ByteImage a(Ix.nx(), Ix.ny(), Ix.nz());
  const int npix = Ix.npix();
  for(int i=0; i<npix; i++)
    a.set(i, image_math::direction(Ix.pixel(i), Iy.pixel(i)));
  return a;
}


Image *stdi::reduce(const Image &img, float r, float scale)
{
  if (img.isNull() || r <=0 ) return NULL;
  FloatImage J = scalespace::smooth(img, scale);
  Image *I = img.clone();
  const int npix = img.npix();
  for(int i=0; i<npix; i++) I->setPixel(i, J(i));

  const int mx = max(1, (int) (img.nx()*r));
  const int my = max(1, (int) (img.ny()*r));
  const int mz = max(1, (int) (img.nz()*r));
  Image *hI = I->scale(mx, my, mz);

  delete I;
  return hI;
}

