/**********************************************************************
 ** image.h
 **********************************************************************/
 
#ifndef IMAGE_H
#define IMAGE_H 1

#include "point.h"
#include "point3d.h"
#include <vector>
#include <iostream>
#include <algorithm>
#include <math.h>
using namespace std;


namespace stdi {

  /**  typedef unsigned char uchar; */
  typedef unsigned char byte;

  /** enum PixelFormat {BYTE, INT, USHORT,SHORT, FLOAT, UINT, RGB, SHORTRGB, INTRGB, FLOATRGB, RGBA, UNKNOWN};  */
  enum PixelFormat {BYTE, INT, USHORT,SHORT, FLOAT, UINT,RGB, SHORTRGB, INTRGB, FLOATRGB, RGBA, CHAR, UNKNOWN};   

  
  enum { Background = 0, Foreground = 255 };
  
  enum NeighborhoodType { Four = 4, Eight = 8 };
  
  /** This structure is used in image class for reference counting */
  struct ArrayData {    
    ArrayData();
    ~ArrayData();
    
    /** increase reference count */
    void ref();
    
    /**
       decrease reference count
       \return true-no reference exists; false-otherwise
     */
    bool deref();
    
    int count; ///< Number of references
    byte *data; ///< image data 
  };

  /* 
     RGB template. Instantiated clases include:
     ::RGB, ::SHORTRGB, ::INTRGB, and ::FLOATRGB
  */
  
  template < typename T >
    struct GenericRgb {
      
      typedef T value_type;
      
      GenericRgb( float intensity ):red(intensity), green(intensity), blue(intensity){}
      
      GenericRgb():red(0),green(0),blue(0){}
      
      GenericRgb( value_type r, value_type g, value_type b ):red(r),green(g),blue(b){}
      
      operator float()
      {return 0.212671*red + 0.715160*green + 0.072169*blue;}
      
      bool operator==(const GenericRgb<T>& rgb){
	if( red == rgb.red && green == rgb.green && blue == rgb.blue ) return true;
	return false;
      }
      value_type blue, green, red;
    };
  
  typedef GenericRgb < unsigned char > Rgb;
  typedef GenericRgb < short > ShortRgb;
  typedef GenericRgb < int > IntRgb;
  typedef GenericRgb < float > FloatRgb;
  
  /*
    RGBA
   */
  struct Rgba : public Rgb{
      
      Rgba( float intensity ):Rgb(intensity),alpha(0){}
      
      Rgba():Rgb(),alpha(0){}
      
      Rgba( byte r, byte g, byte b, byte a ):Rgb(r,g,b),alpha(a){}

      unsigned char alpha;

      operator float()
	  {return 0.212671*red + 0.715160*green + 0.072169*blue;} 
  };

  /** Abstract base class "Image" provides an interface for template "image" */
  class Image {
    
  public:
        
    /**
       Constructor
    */
    Image();
    
    /**
       Constructor
       \param xsize nx
       \param ysize ny
       \param zsize nz
       \param csize nc
       \param tsize nt
     */
    Image(int xsize, int ysize, int zsize, int csize = 1, int tsize = 1 ); 
    
    /** All the pointers returned by the following functions must be deleted after using */
	
    virtual Image *copy() const=0;    /** Shallow copy */
    virtual Image *roi(const Point3D & Pmin, const Point3D & Pmax) const=0;    /** Get the roi image */
    virtual Image *clone() const=0;/** Deep copy */
    virtual Image *frame(int i) const=0; /** Deep copy of the i-th frame */
    virtual Image *channel(int i, int j=0) const = 0; /** Deep copy of the i-th channel of j-th time point */
    virtual Image *timepoint(int i) const =0; /** Deep copy of the i-th time point */
    virtual Image *scale(int mx=-1, int my=-1, int mz=-1) const=0; /** scale image */
    virtual ~Image()=0; /** destructor */
    
    virtual byte *data() const=0; /** Get the pointer of the image data */
    virtual PixelFormat format() const=0; /** Get the format of the image */
    virtual int pixelSize() const=0; /** Get the pixel size of the image */
    
    bool isNull() const; /** Judge if the image is empty */
    int nx() const { return n_x; }/** Get nx */ 
    int ny() const { return n_y; }/** Get ny */ 
    int nz() const { return n_z; }/** Get the total slices of the image: n_z = n_zval * n_c * n_t */
    int nc() const { return n_c; }/** Get nc */
    int nt() const { return n_t; }/** Get nt */
    int nzval() const { return n_zval; }/** Get the real z val */
    int npix() const { return n_pix; }/** Get the total number of pixels */
    bool equalSize(const Image &a)const;/** Judge if two images are equal size */
    int pos(int x, int y, int z) const {return x+y*n_x+z*n_xy;}/** Get the pointer pointing to (x,y,z) in the data pool, here z is for slice */
    int pos(int x, int y) const {return x+y*n_x;}  /** Get the pointer pointing to (x,y) in the data pool */
    
    
    bool inImage(int x, int y, int z) const { return x>=0&&x<n_x&&y>=0&&y<n_y&&z>=0&&z<n_z; }/** Judge if (x,y,z) is within the range of image */
    bool inImage(int x, int y) const {return x>=0&&x<n_x&&y>=0&&y<n_y;}/** Judge if (x,y) is within the range of image */
    bool inImage(const Point &pt) const {return inImage(pt.x(), pt.y()); }/** Judge if pt is within the range of image */
    bool inImage(const Point3D &pt) const {return inImage(pt.x(), pt.y(), pt.z()); }/** Judge if pt is within the range of image */
    bool inImage(int loc) const {return loc>=0&&loc<n_pix;}/** Judge if loc is within the range of image */
     
    
    virtual float pixel(int x, int y, int z, int c, int t = 0) const=0;/** Get pixel value at (x,y,z,c,t), where, z is the real z stack value */
    virtual float pixel(int x, int y, int z) const=0;/** Get pixel value at (x,y,z), where z is the slice value, which means it is not limited by channel or time point */
    virtual float pixel(int x, int y) const=0;/** Get pixel value at (x,y) */
    virtual float pixel(const Point &pt) const=0; /** Get pixel value at pt */
    virtual float pixel(int x) const=0; /** Get pixel value at x */
    virtual void  setPixel(int x, float v)=0; /** Set pixel value at x to be v */
    virtual void  setPixel(int x, int y, float v)=0;/** Set pixel value at (x,y) to be v */
    virtual void  setPixel(int x, int y, int z, float v)=0;/** Set pixel value at (x,y,z) to be v */
    virtual void  setPixel(int x, int y, int z, int c, float v)=0;/** Set pixel value at (x,y,z,c) to be v */
    virtual void  setPixel(int x, int y, int z, int c, int t, float v)=0;/** Set pixel value at (x,y,z,c,t) to be v */
    
    virtual float fmin() const=0;/** Get the min pixel value of the entire image */
    virtual float fmax() const=0;/** Get the max pixel value of the entire image */
    
    vector<bool> interior() const;    
    static vector<bool> interior(int, int, int);
    static Image *reset(Image *&, Image *); /** reset the first image to be the second image */
    
  protected:

    int n_x, n_y, n_z, n_xy, n_pix, n_c, n_t, n_zval, n_xyz, n_xyzc;

  };
  
  /** templete image */
  template <class T>
  class image: public Image {
    
  public:  

    typedef T value_type;
    typedef T & reference;
    typedef const T & const_reference;
       
    image(int xsize=0, int ysize=1, int zsize=1, int csize = 1, int tsize = 1)
      : Image(xsize, ysize, zsize, csize, tsize),
	pixel_size(sizeof(value_type)),
	shd(0), p(0) {
      const int size = n_pix*pixel_size;
      try {
	shd = newData();
	if ( size==0 ) return;
	shd->data = new  byte[size];
      }  catch (const std::bad_alloc &e) {
	cerr << "Error: out of memory\n";
	exit(1);
      }

      p = (value_type *) shd->data;
      memset(shd->data, 0, size); 
    } 
    
    image(const image &a) 
      : Image() { memcpy(this, &a, sizeof(image));  shd->ref(); }

    image & operator= (const image &a) {    
      a.shd->ref();
      if (shd->deref()) deleteData( shd ); 
      memcpy(this, &a, sizeof(image));
      return *this;     
    } 
   
    virtual ~image() {
      if (shd) 
	if(shd->deref() ) 
	  deleteData(shd);
    }
    
    PixelFormat format() const;
    int pixelSize() const {return pixel_size;} //pixel size
    
    image *copy() const { return new  image (*this); }
    image *roi(const Point3D & Pmin, const Point3D & Pmax) const {             // note, for roi operation, we think the input should have only one channel and one timepoint
      int xmin = std::max(0, Pmin.x()), xmax = std::min(n_x-1, Pmax.x());
      int ymin = std::max(0, Pmin.y()), ymax = std::min(n_y-1, Pmax.y());
      int zmin = std::max(0, Pmin.z()), zmax = std::min(n_z-1, Pmax.z());
      if (xmin>xmax||ymin>ymax||zmin>zmax) return NULL;
      image *img = new  image(xmax-xmin+1, ymax-ymin+1, zmax-zmin+1);   
      if (!img) return NULL;
      int x, y, z;
      for(z = zmin; z <= zmax; z++)
	for(y = ymin; y <= ymax; y++)
	  for(x = xmin; x <= xmax; x++)
	    img->at(x-xmin, y-ymin, z-zmin) = this->at(x, y, z);
      return img;    
    }
    
   
    
    image *clone() const { 
      image *a = new  image (n_x, n_y, n_zval,n_c,n_t );
      memcpy(a->data(), data(), n_pix*pixel_size);
      return a;
    }
    
    image duplicate() const {
      image a(n_x, n_y, n_zval,n_c, n_t );
      memcpy(a.data(), this->data(), n_pix*pixel_size);
      return a;
    }
    
    Image *frame(int i) const {
      if (isNull() || i<0 || i>=n_z) return NULL;
      image *a = new  image (n_x, n_y );
      if (!a) return NULL;
      memcpy(a->data(), this->data()+n_x*n_y*pixel_size*i, n_x*n_y*pixel_size);
      return a;
    }
    
    Image *channel(int i, int j) const{
      
      if( isNull() || i<0 || i>=n_c || j < 0 || j >= n_t ) return NULL;
      image *a = new image(n_x,n_y,n_zval);
      if(!a) return NULL;
      memcpy(a->data(), this->data()+n_x*n_y*pixel_size*(j*n_c*n_zval+i*n_zval), n_x*n_y*n_zval*pixel_size);
      return a;
    }
    Image *timepoint(int i ) const{
      
      if( isNull() || i<0 || i>=n_t ) return NULL;
      image * a = new image(n_x,n_y,n_zval,n_c);
      if(!a) return NULL;
      memcpy(a->data(), this->data()+n_x*n_y*pixel_size*(i*n_c*n_zval), n_x*n_y*n_c*n_zval*pixel_size);
      return a;
    }

    image *scale(int mx=-1, int my=-1, int mz=-1) const { // note, for scale operation, we think the input should have only one channel and one timepoint
      if (mx<0) mx = n_x;
      if (my<0) my = n_y;
      if (mz<0) mz = n_z;    
      image *I = new  image(mx, my, mz);  
      int x, y, z;
      for(z=0; z<mz; z++)
	for(y=0; y<my; y++)
	  for(x=0; x<mx; x++)
	    I->at(x, y, z) = this->at(x*n_x/mx, y*n_y/my, z*n_z/mz);    
      return I;
    }

    /** Data pointers */
    byte *data() const { return (byte *) p; }
    value_type *ptr() const {return p;}
    
    /** Pixels operation */
    reference operator() (int x, int y, int z, int c, int t = 0)  {return p[x+y*n_x+z*n_xy+c*n_xyz+t*n_xyzc];}  /** the reason we do not give c a default value is to keep the following 's operation ability across channel and time point */
    reference operator() (int x, int y, int z)  {return p[x+y*n_x+z*n_xy];} 
    reference operator() (int x, int y)  {return p[x+y*n_x];}
    reference operator() (const Point &pt)  {return p[pt.x()+pt.y()*n_x];}
    reference operator() (int x)  {return p[x];} 
    const_reference operator() (int x, int y, int z, int c, int t = 0 ) const {return p[x+y*n_x+z*n_xy+c*n_xyz+t*n_xyzc];}
    const_reference operator() (int x, int y, int z) const {return p[x+y*n_x+z*n_xy];}
    const_reference operator() (int x, int y) const {return p[x+y*n_x];}
    const_reference operator() (const Point &pt) const {return p[pt.x()+pt.y()*n_x];}
    const_reference operator() (int x) const {return p[x];} 

    reference at(int x, int y, int z, int c, int t = 0)  {return p[x+y*n_x+z*n_xy + c*n_xyz + t*n_xyzc];}
    reference at(int x, int y, int z)  {return p[x+y*n_x+z*n_xy];}
    reference at(int x, int y)  {return p[x+y*n_x];}
    reference at(const Point &pt)  {return p[pt.x()+pt.y()*n_x];}
    reference at(int x)  {return p[x];} 
    const_reference at(int x, int y, int z, int c, int t = 0)  const {return p[x+y*n_x+z*n_xy + c*n_xyz + t*n_xyzc];}
    const_reference at(int x, int y, int z) const {return p[x+y*n_x+z*n_xy];}
    const_reference at(int x, int y) const {return p[x+y*n_x];}
    const_reference at(const Point &pt) const {return p[pt.x()+pt.y()*n_x];}
    const_reference at(int x) const {return p[x];} 

    void set(int x, const value_type &v) {p[x] = v;}
    void set(int x, int y, const value_type &v) {p[x+y*n_x] = v;}
    void set(int x, int y, int z, const value_type &v) {p[x+y*n_x+z*n_xy] = v;}
    void set(int x, int y, int z, int c, const value_type &v) {p[x+y*n_x+z*n_xy+c*n_xyz] = v;}
    void set(int x, int y, int z, int c, int t, const value_type &v) {p[x+y*n_x+z*n_xy+c*n_xyz+t*n_xyzc] = v;}
    
    /** Floating point operation */
    float pixel(int x, int y, int z, int c, int t = 0) const { return p[x+y*n_x+z*n_xy+c*n_xyz+t*n_xyzc]; }
    float pixel(int x, int y, int z) const { return p[x+y*n_x+z*n_xy]; }
    float pixel(int x, int y) const {return p[x+y*n_x]; }
    float pixel(const Point &pt) const {return p[pt.x() + pt.y()*n_x]; }
    float pixel(int x) const { return p[x]; }
    void  setPixel(int x, float v) { p[x] = (value_type)v; }
    void  setPixel(int x, int y, float v) { p[x+y*n_x] = (value_type)v; }
    void  setPixel(int x, int y, int z, float v) { p[x+y*n_x+z*n_xy] = (value_type)v; }
    void  setPixel(int x, int y, int z, int c, float v) { p[x+y*n_x+z*n_xy+c*n_xyz] = (value_type)v; }
    void  setPixel(int x, int y, int z, int c, int t, float v) { p[x+y*n_x+z*n_xy+c*n_xyz+t*n_xyzc] = (value_type)v; }
    
    value_type min() const {
      if (isNull()) return value_type();   
      return *min_element(p, p+n_pix);
    }

    value_type max() const {
      if (isNull()) return value_type();
      return *max_element(p, p+n_pix);
    }

    void reset(const value_type &v) const {
      if (pixel_size==1) memset(this->p, int(v), n_pix);
      else for(int k=0; k<n_pix; k++) p[k] = v;
    }
    
    float fmin() const { return min(); }
    float fmax() const { return max(); }

    /** Scanlines(2D) 
     for line operation, we think the input should have only one channel and one timepoint
    */
    void lineX(int y, value_type *line) const {
      memcpy(line, p+y*n_x, n_x*pixel_size);
    }
    void lineY(int x, value_type *line) const {
      value_type *q = p+x;
      for(int y=0; y<n_y; y++) line[y] = q[y*n_x];
    }
    void setLineX(int y, const value_type *line) {
      value_type *q = p+y*n_x;
      for(int x=0; x<n_x; x++) q[x] = line[x];
    }
    void setLineY(int x, value_type *line) {
      value_type *q = p+x;
      for(int y=0; y<n_y; y++)  q[y*n_x] = line[y];
    }

    /** Scanlines(3D) */
    void lineX(int y, int z, value_type *line) const {
      memcpy(line, p+z*n_xy+y*n_x, n_x*pixel_size);
    }
    void lineY(int x, int z, value_type *line) const {
      value_type *q = p+x+z*n_xy;
      for(int y=0; y<n_y; y++)  line[y] = q[y*n_x];
    }
    void lineZ(int x, int y, value_type *line) const {
      value_type *q = p+x+y*n_x;
      for(int z=0; z<n_z; z++) line[z] = q[z*n_xy];
    }
    void setLineX(int y, int z, const value_type *line) {
      value_type *q = p+z*n_xy+y*n_x;
      for(int x=0; x<n_x; x++) q[x] = line[x];
    }   
    void setLineY(int x, int z, value_type *line) {
      value_type *q = p+x+z*n_xy;
      for(int y=0; y<n_y; y++)  q[y*n_x] = line[y];
    }
    void setLineZ(int x, int y, value_type *line) {
      value_type *q = p+x+y*n_x;
      for(int z=0; z<n_z; z++) q[z*n_xy] = line[z];
    }


  protected:
    
    int pixel_size;
    ArrayData *shd; ///< Shared array data
    value_type *p; 
    
    ArrayData * newData() { return new  ArrayData; }
    void deleteData( ArrayData * & d ) { delete d; d = 0;}

  };


  template<> inline PixelFormat image<byte>::format() const {return BYTE;}
  template<> inline PixelFormat image<short>::format() const {return SHORT;}
  template<> inline PixelFormat image<int>::format() const {return INT;}
  template<> inline PixelFormat image<float>::format() const {return FLOAT;}
  template<> inline PixelFormat image<unsigned short>::format() const {return USHORT;}
  template<> inline PixelFormat image<unsigned int>::format() const {return UINT;}
  template<> inline PixelFormat image<Rgb>::format() const {return RGB;}
  template<> inline PixelFormat image<IntRgb>::format() const {return INTRGB;}
  template<> inline PixelFormat image<ShortRgb>::format() const {return SHORTRGB;}
  template<> inline PixelFormat image<FloatRgb>::format() const {return FLOATRGB;}
  template<> inline PixelFormat image<Rgba>::format() const {return RGBA;}
  template<> inline PixelFormat image<char>::format() const {return CHAR;}

  /** Type definitions   */
  typedef image<byte> ByteImage;
  typedef image<short> ShortImage;
  typedef image<int> IntImage;
  typedef image<float> FloatImage;  
  typedef Image * ImagePtr;
  typedef image<unsigned short> UShortImage;
  typedef image<unsigned int> UIntImage; 
  typedef image<Rgb> RgbImage; 
  typedef image<IntRgb> IntRgbImage;
  typedef image<ShortRgb> ShortRgbImage;
  typedef image<FloatRgb> FloatRgbImage;
  typedef image<Rgba> RgbaImage;
  typedef image<char> CharImage;

  /** Format conversion */
  FloatImage toFloatImage(const Image &);
  RgbImage toRgbImage(const Image & );
  ByteImage toByteImage(const Image &img, bool forced_scaling);

  class Roi {
    
  public:
    
    typedef int coord_type;
    
    Roi(int x_min, int y_min, int x_max, int y_max);
    Roi(const PointArray &boundary);
    Roi();
    ~Roi();
    
    /** Get/set properties */
    coord_type xmin() const;
    coord_type ymin() const; 
    coord_type xmax() const; 
    coord_type ymax() const;  
    void set(int x_min, int y_min, int x_max, int y_max);    
    
    bool isEmpty() const;

    /** ROI operations */
    Image *getSubImage(const Image &) const;    
    static Image *getSubImage(const Image &, const Roi &);
    static Image *getSubImage(const Image &, 
			      int x_min, int y_min, int x_max, int y_max);
    
  private:

    coord_type _xmin, _ymin, _xmax, _ymax;

  };

  /** Misc. functions for code compatibility */
  Image *reduce(const Image &img, 
		float ratio=0.5,
		float smooth_scale = 1.0);

  void out_of_memory();///< Global new handler
  int round_double(double x);
   
} // stdi

#endif // IMAGE_H
